from surety import Array, Set, String


def test_array_no_generate_value():
    assert Array(String).generate_value() is None


def test_array_generate():
    array = Array(String)
    array.generate()
    assert isinstance(array.value, list)


def test_array_generate_value():
    value = Array(String).value
    assert isinstance(value, list)


def test_len_of_array():
    generated_array = Array(String, max_len=3, min_len=3)
    assert len(generated_array) == 3


def test_set_array_value():
    generated_array = Array(String, min_len=1, max_len=1)
    generated_array[0] = String(default='test')
    assert generated_array.value == ['test']


def test_iterate_by_array():
    generated_array = Array(String(default='test'), min_len=1, max_len=1)
    result = []

    for item in generated_array:
        result.append(item.value)

    assert result == ['test']


def test_append():
    generated_array = Array(String(default='test'), min_len=1, max_len=1)
    generated_array.append(String(default='test2'))
    assert generated_array.value == ['test', 'test2']


def test_with_values():
    generated_array = Array(String).with_values([
        'test1', 'test2', 'test3'
    ])
    assert generated_array.value == ['test1', 'test2', 'test3']


def test_with_no_values():
    array = Array(
        String(default='test'), min_len=1, max_len=1
    )
    array.with_values('')
    assert array.value == []


def test_set_with_no_values():
    new_set = Set(
        String(default='test'), min_len=1, max_len=1
    )
    new_set.with_values('')
    assert new_set.value == set()



import pytest

from surety.process.dictionary import apply_modifier, dict_predicate


def test_use_function_name_as_repr():
    @dict_predicate
    def filter_dict_function():
        pass

    assert repr(filter_dict_function) == 'filter_dict_function'


def test_use_custom_name_as_repr():
    @dict_predicate(name='test_filter')
    def filter_dict_function():
        pass

    assert repr(filter_dict_function) == 'test_filter'


def test_function_with_name_is_callable():
    @dict_predicate(name='test_f')
    def filter_dict_function():
        return True

    assert filter_dict_function() is True


def test_apply_none_modifier():
    with pytest.raises(ValueError,
                       match='modifier is not dictionary nor callable'):
        apply_modifier({}, None)


def test_apply_list_modifier_is_empty():
    with pytest.raises(ValueError,  match='list modifier is empty'):
        apply_modifier({}, [])


def test_apply_list_modifier_to_not_list():
    with pytest.raises(ValueError,  match=(
            "cannot apply list modifier to <class 'dict'> type")):
        apply_modifier({'a': 1}, ['test'])



from tests.data import TypeOne


def test_saved_kwargs_on_generate():
    type_one = TypeOne()
    assert type_one._kwargs == {  # pylint: disable=protected-access
        'allow_none': False,
        'default': None,
        'name': None,
        'required': True
    }


def test_do_not_override_kwargs():
    type_one = TypeOne()
    type_one.save_kwargs({'new_value': 'test'})

    assert type_one._kwargs == {  # pylint: disable=protected-access
        'allow_none': False,
        'default': None,
        'name': None,
        'required': True
    }


def test_generate_with_default():
    default_value = 'Default value'
    type_one = TypeOne(default=default_value)

    assert type_one.value == default_value


def test_generate_with_default_function():
    default_value = 'Generated by function'
    type_one = TypeOne(default=lambda: default_value)

    assert type_one.value == default_value


def test_repr_of_generated_field():
    type_one = TypeOne()
    assert repr(type_one) == type_one.value



from surety.sdk import fakeable
from surety.sdk.fakeable import (
    Fakeable, fake_string_attr, generate_float, generate_string,
)


def test_generate_string_fixed_size():
    value = generate_string(size=28)
    assert len(value) == 28


def test_fake_attribute():
    value = fake_string_attr(Fakeable.CompanyEmail)
    assert '@' in value


def test_max_len_in_fake_attrs():
    value = fake_string_attr(Fakeable.Uuid, max_len=10)
    assert len(value) == 10


def test_generate_float_i_len():
    value = generate_float(i_len=3)
    assert len(str(value).split('.', maxsplit=1)[0]) <= 3


def test_generate_float_f_len():
    value = generate_float(f_len=5)
    assert len(str(value).split('.')[1]) <= 5


def test_generate_float_fixed_f_len(monkeypatch):
    monkeypatch.setattr(
        fakeable.random,
        'randint',
        lambda x, y: 1 if y < 10 else 10
    )
    value = generate_float(fixed_f_len=True, f_len=2, i_len=0,
                           integer_allowed=False)
    assert len(str(value).split('.')[1]) == 2


def test_fakeable_to_list():
    assert len(Fakeable.to_list()) == 81


def test_fakeable_include():
    assert Fakeable.AmPm in Fakeable.to_list()


def test_fakeable_exclude():
    assert len(Fakeable.to_list(exclude=[
        Fakeable.AmPm, Fakeable.Address, Fakeable.CompanyEmail
    ])) == 78


def test_fakeable_exclude_value():
    assert Fakeable.DateTime not in Fakeable.to_list(exclude=Fakeable.DateTime)



from surety import Array
from tests.data import Base, ComposeBase, Mix, Optional, TypeOne


def test_field_generated():
    assert TypeOne().generated


def test_embedded_field_generated():
    assert Base().IntField.generated


def test_optional_field_not_generated():
    assert not Optional().OptString.generated


def test_optional_field_generated_in_full():
    assert Optional(is_full=True).OptString.generated


def test_dict_generated():
    assert ComposeBase().generated


def test_embedded_dict_generated():
    assert ComposeBase().FirstBase.generated


def test_embedded_optional_dict_not_generated():
    assert not ComposeBase().OptBase.generated


def test_embedded_optional_dict_generated_in_full():
    assert ComposeBase(is_full=True).OptBase.generated


def test_array_generated():
    assert Array(Base).generated


def test_array_dict_generated():
    assert Array(Base)[0].generated


def test_array_embedded_field_generated():
    assert Array(Base)[0].StringField.generated


def test_array_optional_not_generated():
    assert not Array(ComposeBase)[0].OptBase.generated


def test_array_optional_generated_in_full():
    assert Array(ComposeBase, is_full=True)[0].OptBase.generated


def test_optional_array_not_generated():
    assert not Mix().OptFullBaseArray.generated


def test_optional_array_generated_in_full():
    assert Mix(is_full=True).OptFullBaseArray.generated


def test_optional_allownone_not_generated():
    assert not Mix().NotNoneAllowedNone.OptNoneInt.generated


def test_optional_allownone_generated_in_full():
    assert Mix(is_full=True).NotNoneAllowedNone.OptNoneInt.generated



from surety.sdk.base_enum import BaseEnum


def test_base_enum_single_value_to_list():
    class SingleVal(BaseEnum):
        First = 'first'

    assert SingleVal.to_list() == ['first']


def test_base_enum_multiple_values_to_list():
    class MultiVal(BaseEnum):
        First = 'first'
        Second = 'second'

    assert MultiVal.to_list() == ['first', 'second']


def test_base_enum_skip_others_in_to_list():
    class ExtrasVal(BaseEnum):
        First = 'first'
        Second = 'second'

        @property
        def prop_first(self):
            return None

        def method_first(self):
            return self

    assert ExtrasVal.to_list() == ['first', 'second']


def test_base_enum_exclude():
    class Values(BaseEnum):
        First = 'first'
        Second = 'second'
        Third = 'third'
        Last = 'last'

    assert sorted(Values.to_list(exclude=Values.Third)) == sorted(
        ['first', 'second', 'last']
    )


def test_base_enum_exclude_multiple():
    class Values(BaseEnum):
        First = 'first'
        Second = 'second'
        Third = 'third'
        Last = 'last'

    assert sorted(Values.to_list(
        exclude=[Values.Third, Values.First]
    )) == sorted(['second', 'last'])



import inspect

from collections.abc import Mapping


def filter_dict(function_or_value, dict_to_filter):
    """
    Filter by value
      >>> filter_dict(123, {'a': 123, 'b': 1234})
      {'b': 1234}

    Filter by value not applicable
      >>> filter_dict(123, {'a': 1234, 'b': 5123})
      {'a': 1234, 'b': 5123}

    Embedded filter by value
      >>> filter_dict(123, {'a': {'c': 123}, 'b': 1234})
      {'b': 1234}

    Embedded with extra by value
      >>> filter_dict(123, {'a': {'c': 123, 'd': 432}, 'b': 1234})
      {'a': {'d': 432}, 'b': 1234}

    Embedded mixed filter
      >>> filter_dict(123, {'a': {'c': 123, 'd': 432}, 'b': 123, 'e': 'test'})
      {'a': {'d': 432}, 'e': 'test'}

    Filter by callable
      >>> filter_dict(lambda x: x % 2 == 0, {'a': 532, 'b': 891})
      {'a': 532}

    Filter by callable not applicable
      >>> filter_dict(lambda x: x % 2 == 0, {'a': 538, 'b': 8})
      {'a': 538, 'b': 8}

    Embedded filter by callable
      >>> filter_dict(lambda x: bool(x), {'a': {'c': False}, 'b': 'test'})
      {'b': 'test'}

    Embedded with extra by callable
      >>> filter_dict(
      ...   lambda x: 'a' in x, {'a': {'c': 'ba', 'd': 'tt'}, 'b': 'd'})
      {'a': {'c': 'ba'}}

    Embedded mixed filter
      >>> filter_dict(
      ...   lambda x: bool(x), {'a': {'c': True, 'd': 0}, 'b': 'test', 'e': []}
      ...   )
      {'a': {'c': True}, 'b': 'test'}

    """
    func = function_or_value

    if not callable(function_or_value):
        def new_func(value):
            return value != function_or_value

        func = new_func

    result = {}

    for key, value in dict_to_filter.items():
        if isinstance(value, Mapping):
            value = filter_dict(func, value)

            if value:
                result[key] = value

        elif func(value):
            result[key] = value

    return result


def exclude_none_from_kwargs(kwargs):
    """
    Remove single None value
      >>> exclude_none_from_kwargs({'a': None, 'b': 1234})
      {'b': 1234}

    Remove multiple None value
      >>> exclude_none_from_kwargs({'a': None, 'b': 1234, 'c': None})
      {'b': 1234}

    Remove all None values
      >>> exclude_none_from_kwargs({'a': None, 'c': None})
      {}

    Do nothing when there is no None value
      >>> exclude_none_from_kwargs({'a': 123, 'b': 1234})
      {'a': 123, 'b': 1234}
    """
    return {k: v for k, v in kwargs.items() if v is not None}


def is_callable_with_strict_args(func, args_count=1):
    if not callable(func):
        return False

    if hasattr(func, 'func'):
        func = func.func

    if not inspect.isfunction(func):
        raise TypeError(f'Expected function, got {type(func)}')

    argspec = inspect.getfullargspec(func)

    if argspec.varkw is None and len(argspec.args) == args_count:
        return True

    raise TypeError(f'Expected callable with {args_count} args. '
                    f'Got: "{func.__name__}" func with {argspec}')



import random

from surety.sdk.field import Field


class Array(Field):
    def __init__(self, field, name=None, required=True, allow_none=False,
                 is_full=False, min_len=1, max_len=1):
        if not hasattr(self, '_kwargs'):
            self.save_kwargs(locals())

        self.is_full = is_full
        self._field = field
        self.min_len = min_len
        self.max_len = max_len
        self._value = None
        super().__init__(name, required, allow_none)

    @property
    def field(self):
        if not isinstance(
                self._field, Field) and issubclass(self._field, Field):
            self._field = self._field()

        return self._field

    @property
    def value(self):
        if self._value is None:
            return None

        return [field.value for field in self._value]

    @property
    def full_value(self):
        if self._value is None:
            return []

        return [field.full_value for field in self._value]

    def generate_value(self):
        pass

    def _generate(self, is_full=False, with_data=True, required=True,
                  use_default=None):
        # Not processed: duplicated entities, duplicated empty dicts
        if with_data and (is_full or required):
            self._value = [self.field(is_full=is_full) for _ in range(
                random.randint(self.min_len, self.max_len)
            )]

    def generate(self, is_full=False):  # pylint: disable=arguments-renamed
        self._generate(is_full=is_full)

    def __len__(self):
        return len(self._value)

    def __getitem__(self, item):
        return self._value[item]  # Return Field to provide attribute access

    def __setitem__(self, key, value):
        self._value[key] = value

    def __iter__(self):
        return iter(self._value)

    def append(self, value):
        self._value.append(value)

    def with_values(self, values):
        if isinstance(values, list):
            _values = []

            for value in values:
                if not isinstance(value, self.field.__class__):
                    value = self.field(is_full=False).with_values(value)
                _values.append(value)
            self._value = _values

            return self

        return super().with_values(values)


class Set(Array):
    def with_values(self, values):
        if isinstance(values, (set, list)):
            _values = set()

            for value in values:
                if not isinstance(value, self.field.__class__):
                    value = self.field(is_full=False).with_values(value)
                _values.add(value)
            self._value = _values

            return self

        return super().with_values(values)

    @property
    def value(self):
        if self._value is None:
            return None

        return {field.value for field in self._value}

    @property
    def full_value(self):
        if self._value is None:
            return []

        return {field.full_value for field in self._value}



from tests.data import Mix


def test_dict_is_not_allowed_none():
    assert not Mix.NotNoneAllowedNone.allow_none


def test_embedded_dict_is_allowed_none():
    assert Mix.NoneAllowedNone.allow_none


def test_embedded_in_dict_is_allowed_none():
    assert Mix.NoneAllowedNone.NoneString.allow_none


def test_array_none_is_allowed():
    assert Mix.NoneTypeArray.allow_none


def test_embedded_array_none_is_allowed():
    assert Mix.NoneTypeArray.field.NoneInt.allow_none



import json

import pytest

from surety import Array, Dictionary, String


def test_dictionary_no_generate_value():
    assert Dictionary().generate_value() is None


def test_dictionary_with_values_with_field():
    class TempDict(Dictionary):
        First = String(name='first')
        Second = String(name='second')

    temp_dict = TempDict().with_values({
        TempDict.First: 'first_val',
        TempDict.Second: 'second_val'
    })
    assert temp_dict.value == {'first': 'first_val', 'second': 'second_val'}


def test_dictionary_with_values_no_field():
    class TempDict(Dictionary):
        First = String(name='first')

    with pytest.raises(AttributeError, match='No attribute with name second'):
        TempDict().with_values({'second': 'second_val'})


def test_dictionary_with_values_skip_nones():
    class TempDict(Dictionary):
        First = String(name='first', default='First')
        Second = String(name='second')

    temp_dict = TempDict().with_values({
        TempDict.First.name: None,
        TempDict.Second.name: 'second_val'
    })
    assert temp_dict.value == {'first': 'First', 'second': 'second_val'}


def test_dictionary_with_values_set_to_field():
    class TempDict(Dictionary):
        First = String(name='first')

    temp_dict = TempDict().with_values({
        TempDict.First.name: String(default='test'),
    })
    assert temp_dict.value == {'first': 'test'}


def test_dictionary_set_with_json():
    class TempDict(Dictionary):
        First = String(name='first')

    class OuterDict(Dictionary):
        Dict = TempDict(name='dict')

    outer_dict = OuterDict().with_values({
        OuterDict.Dict.name: json.dumps({TempDict.First.name: 'test'})
    })
    assert outer_dict.value == {'dict': {'first': 'test'}}


def test_dictionary_set_with_bytes():
    class TempDict(Dictionary):
        First = String(name='first')

    class OuterDict(Dictionary):
        Dict = TempDict(name='dict')

    outer_dict = OuterDict().with_values({
        OuterDict.Dict.name: json.dumps({
            TempDict.First.name: 'test1'
        }).encode('utf-8')
    })
    assert outer_dict.value == {'dict': {'first': 'test1'}}


def test_dictionary_set_with_string():
    class TempDict(Dictionary):
        First = String(name='first')

    class OuterDict(Dictionary):
        Dict = TempDict(name='dict')

    with pytest.raises(AttributeError, match='Assigning entity to primitive'):
        OuterDict().with_values({OuterDict.Dict.name: 'test'})


def test_dictionary_of_arrays():
    class TempDict(Dictionary):
        Array = Array(String, name='arr')

    temp_dict = TempDict().with_values({
        TempDict.Array.name: [
            String(default='test'),
            String(default='test2')
        ]
    })
    assert temp_dict.value == {'arr': ['test', 'test2']}



import pytest

from surety.process.common import is_callable_with_strict_args


def test_arg_not_callable():
    assert is_callable_with_strict_args('test') is False


def test_single_arg_function():
    assert is_callable_with_strict_args(func=lambda x: x)


def test_multiple_args_function_match():
    assert is_callable_with_strict_args(func=lambda x, y: x + y, args_count=2)


def test_func_as_a_class_property():
    class Temp:
        @staticmethod
        def func(arg):
            return arg

    assert is_callable_with_strict_args(Temp)


def test_class_property_not_a_function():
    class Temp:
        func = None

    with pytest.raises(TypeError):
        is_callable_with_strict_args(Temp)


def test_arg_count_mismatch():
    with pytest.raises(TypeError):
        is_callable_with_strict_args(func=lambda: True)



from tests.data import Base, Mix, Optional


def test_required_fields_with_values():
    new_value = 'my_new_value'
    entity = Base().with_values({Base.StringField.name: new_value})
    assert entity.StringField.value == new_value


def test_optional_field_with_values_generated():
    entity = Optional().with_values({Optional.OptString.name: 'test_value'})
    assert entity.OptString.generated


def test_optional_field_with_values_new_value():
    new_value = 'optional_string'
    entity = Optional().with_values({Optional.OptString.name: new_value})
    assert entity.OptString.value == new_value


def test_embedded_dict_with_values_dict_generated():
    entity = Mix().with_values({
        Mix.OptBase.name: {Base.StringField.name: 'test_string'},
    })
    assert entity.OptBase.generated


def test_embedded_dict_with_values_field_generated():
    entity = Mix().with_values({
        Mix.OptBase.name: {Base.StringField.name: 'test_val'},
    })
    assert entity.OptBase.StringField.generated


def test_embedded_dict_with_values_field_not_affect_other_generated():
    entity = Mix().with_values({
        Mix.OptBase.name: {Base.StringField.name: 'value'},
    })
    assert not entity.OptBase.IntField.generated


def test_embedded_dict_with_values_new_value():
    new_value = 'embedded_dict_string'
    entity = Mix().with_values({
        Mix.OptBase.name: {Base.StringField.name: new_value}
    })
    assert entity.OptBase.StringField.value == new_value


def test_embedded_dict_with_values_new_dictionary_value():
    new_value = 'new_dict_string'
    entity = Mix().with_values({
        Mix.OptBase.name: {Base.StringField.name: new_value}
    })
    assert entity.OptBase.value == {Base.StringField.name: new_value}


def test_array_of_fields_with_values_is_generated():
    entity = Mix().with_values({
        Mix.OptTypeArray.name: ['random_str1', 'random_str2'],
    })
    assert entity.OptTypeArray.generated


def test_set_of_fields_with_values_is_generated():
    entity = Mix().with_values({
        Mix.OptTypeSet.name: {'random_str1', 'random_str2'},
    })
    assert entity.OptTypeSet.generated


def test_array_of_fields_with_values_value():
    new_values = ['str1', 'str2', 'str3']
    entity = Mix().with_values({Mix.OptTypeArray.name: new_values})
    assert entity.OptTypeArray.value == new_values


def test_set_of_fields_with_values_value():
    new_values = {'str1', 'str2', 'str3'}
    entity = Mix().with_values({Mix.OptTypeSet.name: new_values})
    assert entity.OptTypeSet.value == new_values


def test_array_of_fields_with_values_override_value():
    new_values = ['r_str1', 'r_str2', 'r_str3']
    entity = Mix().with_values({Mix.ReqTypeArray.name: new_values})
    assert entity.ReqTypeArray.value == new_values


def test_set_of_fields_with_values_override_value():
    new_values = {'r_str1', 'r_str2', 'r_str3'}
    entity = Mix().with_values({Mix.ReqTypeSet.name: new_values})
    assert entity.ReqTypeSet.value == new_values


def test_array_of_dicts_with_values_is_generated():
    entity = Mix().with_values({Mix.OptBaseArray.name: [
        {Base.StringField.name: 'randoms_name1'},
        {Base.StringField.name: 'randoms_name2'},
    ]})
    assert entity.OptBaseArray.generated


def test_array_of_dicts_with_value_is_generated():
    entity = Mix().with_values({Mix.OptBaseArray.name: [
        {Base.StringField.name: 'name1', Base.IntField.name: 1},
        {Base.StringField.name: 'name2', Base.IntField.name: 2},
    ]})
    assert entity.OptBaseArray.value == [
        {'string_field': 'name1', 'int_field': 1},
        {'string_field': 'name2', 'int_field': 2}
    ]



import decimal
import random

from faker import Faker

from surety.sdk.base_enum import BaseEnum


class Fakeable(BaseEnum):
    Address = 'address'
    AmPm = 'am_pm'
    AndroidToken = 'android_platform_token'
    BankCountry = 'bank_country'
    Boolean = 'boolean'
    BuildingNumber = 'building_number'
    City = 'city'
    CityPrefix = 'city_prefix'
    CitySuffix = 'city_suffix'
    ColorName = 'color_name'
    SafeColorName = 'safe_color_name'
    Company = 'company'
    CompanyEmail = 'company_email'
    Coordinate = 'coordinate'
    Country = 'country'
    CountryCode = 'country_code'
    CreditCardExpire = 'credit_card_expire'
    CreditCardFull = 'credit_card_full'
    CreditCardNumber = 'credit_card_number'
    CreditCardProvider = 'credit_card_provider'
    CreditCardSecurityCode = 'credit_card_security_code'
    Currency = 'currency'
    CurrencyCode = 'currency_code'
    CurrencyName = 'currency_name'
    Date = 'date'
    DateOfBirth = 'date_of_birth'
    DateThisMonth = 'date_this_month'
    DateTime = 'date_time'
    DayOfMonth = 'day_of_month'
    DomainName = 'domain_name'
    Email = 'email'
    FileExtension = 'file_extension'
    FileName = 'file_name'
    FirstName = 'first_name'
    FutureDate = 'future_date'
    FutureDateTime = 'future_datetime'
    HexColor = 'hex_color'
    Hostname = 'hostname'
    Iban = 'iban'
    ImageUrl = 'image_url'
    IosToken = 'ios_platform_token'
    LastName = 'last_name'
    Latitude = 'latitude'
    Longitude = 'longitude'
    Month = 'month'
    MonthName = 'month_name'
    Name = 'name'
    Password = 'password'
    PastDate = 'past_date'
    PastDateTime = 'past_datetime'
    PhoneNumber = 'phone_number'
    PostalCode = 'postalcode'
    Decimal = 'pydecimal'
    Dict = 'pydict'
    Float = 'pyfloat'
    Iterable = 'pyiterable'
    List = 'pylist'
    Str = 'pystr'
    Digit = 'random_digit'
    Int = 'random_int'
    Number = 'random_number'
    SecondaryAddress = 'secondary_address'
    Sentence = 'sentence'
    Sentences = 'sentences'
    State = 'state'
    StateAbbr = 'state_abbr'
    StreetAddress = 'street_address'
    StreetName = 'street_name'
    Text = 'text'
    Time = 'time'
    TimeDelta = 'time_delta'
    Timezone = 'timezone'
    UnixTime = 'unix_time'
    Url = 'url'
    UserName = 'user_name'
    Uuid = 'uuid4'
    Word = 'word'
    Year = 'year'
    Zipcode = 'zipcode'
    RgbColor = 'rgb_color'
    RgbCssColor = 'rgb_css_color'


FAKER = Faker()


def generate_string(size=None, min_len=None, max_len=None):
    if size:
        min_len = size
        max_len = size

    return FAKER.pystr(min_chars=min_len, max_chars=max_len or 20)


def fake_string_attr(attr_name, min_len=None, max_len=None):
    if attr_name and hasattr(FAKER, attr_name):
        result = getattr(FAKER, attr_name)()

        if max_len and len(result) > max_len:
            result = result[:max_len]

        return str(result)

    return generate_string(min_len=min_len, max_len=max_len)


def generate_float(i_len=None, f_len=None, integer_allowed=True, max_val=None,
                   fixed_f_len=False, i_min=None):
    """
      Generates random float with floating number of digits in integer and
    fractional part. Fractional part length can be fixed for e.g. money types.
      Cannot be used with f_len=1 & fixed_f_len=True - fraction part will be 0
    :param i_len: max length of integer part
    :param f_len: max length of fraction part
    :param integer_allowed: whether float might end with zero like 21.0
    :type max_val: int or None
    :param max_val: max allowed value, result will be generated up to the value
      specified (never equal)
    :param fixed_f_len: whether length of floating part is fixed
    :param i_min: min value of integer part, default value is 0
    """
    f_len = 6 if f_len is None else f_len
    i_max = max_val or 10 ** (3 if i_len is None else i_len)
    f_max = 10 ** f_len
    f_min = 0 if integer_allowed else 1
    i_min = i_min or 0
    i_part = random.randint(i_min, i_max - 1)
    f_part = random.randint(f_min, f_max - 1)

    if fixed_f_len and f_part % 10 == 0:
        f_part += random.randint(1, 9)

    return float(i_part + f_part / decimal.Decimal(f_max))



from tests.data import Base, ComposeBase, Mix


def test_field_no_value():
    assert Base.StringField.value is None


def test_embedded_dict_no_value():
    assert ComposeBase.FirstBase.value is None


def test_embedded_dict_field_no_value():
    assert ComposeBase.FirstBase.IntField.value is None


def test_array_no_value():
    assert Mix.ReqFullBaseArray.value is None



from abc import ABCMeta, abstractmethod


class Field(metaclass=ABCMeta):
    def __init__(self, name=None, required=True, allow_none=False,
                 default=None):
        self.required = required
        self.allow_none = allow_none
        self.name = name
        self.default = default
        self._value = None

        if not hasattr(self, '_kwargs'):
            self.save_kwargs(locals())

        if not hasattr(self, 'is_full'):
            self.is_full = True

        if name is None:
            self._generate(is_full=self.is_full)
        else:
            self._generate(with_data=False)

    def save_kwargs(self, local_vars):
        if hasattr(self, '_kwargs'):
            return

        self._kwargs = {k: v for k, v in local_vars.items()
                        if k not in ('self', '__class__')}

    @property
    def generated(self):
        if self._value is None:
            return self.required and self.allow_none

        return True

    @property
    def value(self):
        return self._value

    @property
    def full_value(self):
        return self.value

    def __call__(self, is_full, with_data=True):
        new_instance = self.__class__(**self._kwargs)
        new_instance._generate(is_full=is_full, with_data=with_data,
                               required=new_instance.required)

        return new_instance

    def with_values(self, values):
        self._value = values

        return self

    def _generate(self, is_full=True, with_data=True, required=True,
                  use_default=True):
        if not with_data:
            return

        if is_full or (required and not self.allow_none):
            if use_default and self.default is not None:
                if callable(self.default):
                    self._value = self.default()
                else:
                    self._value = self.default
            else:
                self._value = self.generate_value()

    def generate(self, use_default=False):
        self._generate(use_default=use_default)

    @abstractmethod
    def generate_value(self):
        """Valid value generator based on type"""

    def __repr__(self):
        return str(self.value)



from datetime import date, datetime, timedelta, timezone

from surety.sdk import dates


def test_get_current_time_utc():
    current_time = datetime.now(timezone.utc)
    got_time = dates.get_current_time_utc()
    assert got_time - current_time < timedelta(seconds=3)


def test_generate_date():
    got_date = dates.generate_date()
    assert isinstance(got_date, date)


def test_to_timestamp():
    current_time = datetime.now(timezone.utc)
    got_ts = dates.to_timestamp(current_time)
    assert isinstance(got_ts, int)


def test_to_format():
    date_time = datetime.fromisoformat('2023-02-16 21:32')
    got_str = dates.to_format(date_time)
    assert got_str == '2023-02-16T21:32:00Z'

def test_str_to_date():
    got_date = dates.str_to_date('2021-04-21', dates.Pattern.DATE)
    assert isinstance(got_date, datetime)


def test_to_timezone():
    current_time = datetime.now(timezone.utc)
    got_datetime = dates.to_timezone(current_time, 'US/Eastern')
    assert got_datetime.tzinfo.zone == 'US/Eastern'



import random
import string

from surety import Array, Dictionary, Field, Set


class TypeOne(Field):
    def generate_value(self):
        return ''.join(random.sample(string.ascii_letters + string.digits, 6))


class TypeTwo(Field):
    def generate_value(self):
        return random.randint(0, 1000)


class Base(Dictionary):
    StringField = TypeOne(name='string_field')
    IntField = TypeTwo(name='int_field')


class Optional(Dictionary):
    OptString = TypeOne(name='opt_string', required=False)
    OptInt = TypeTwo(name='opt_int', required=False)


class AllowedNone(Dictionary):
    NoneString = TypeOne(name='none_str', allow_none=True)
    NoneInt = TypeTwo(name='none_int', allow_none=True)
    OptNoneInt = TypeTwo(name='opt_none_int', allow_none=True,
                         required=False)


class ComposeBase(Dictionary):
    FirstBase = Base(name='first_base')
    SecondBase = Base(name='second_base')
    ThirdBase = Base(name='third_base')
    OptBase = Base(name='opt_base', required=False)


class ComposeOptional(Dictionary):
    FirstOpt = Optional(name='first_optional')
    OptionalOpt = Optional(name='optional_opt', required=False)


class Mix(Dictionary):
    ReqBase = Base(name='req_base')
    OptBase = Base(name='opt_base', required=False)
    NotNoneAllowedNone = AllowedNone(name='not_allowed')
    NoneAllowedNone = AllowedNone(name='none_allowed', allow_none=True)
    ReqOptional = Optional(name='req_optional')
    OptOptional = Optional(name='opt_optional', required=False)
    ReqTypeArray = Array(TypeOne, name='req_type_array')
    ReqTypeSet = Set(TypeOne, name='req_type_set')
    OptTypeArray = Array(TypeOne, name='opt_type_array', required=False)
    OptTypeSet = Set(TypeOne, name='opt_base_set', required=False)
    NoneTypeArray = Array(AllowedNone, name='none_type_array', allow_none=True)
    ReqBaseArray = Array(Base, name='req_base_array')
    OptBaseArray = Array(Base, name='opt_base_array', required=False)
    ReqFullBaseArray = Array(
        Base(is_full=True), name='req_full_base_array'
    )
    OptFullBaseArray = Array(
        Base(is_full=True), name='opt_full_base_array', required=False
    )



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'Surety'
copyright = '2026, Elena Kulgavaya'
author = 'Elena Kulgavaya'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

templates_path = ['_templates']
exclude_patterns = []



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_static_path = ['_static']
html_css_files = ["custom.css"]
html_favicon = "_static/surety.ico"
html_logo = "_static/surety.png"
html_sidebars = {
    "**": [
        "sidebar/brand.html",
        "sidebar/search.html",
        "sidebar/navigation.html",
        "sidebar-links.html",   # <-- our custom block
    ]
}
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinx_autodoc_typehints",
]

html_theme = "furo"
html_theme_options = {
    "source_repository": "https://github.com/elenakulgavaya/surety/",
    "source_branch": "main",
    "source_directory": "docs/source/",
    "sidebar_hide_name": True,
}

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}



from tests.data import Base, ComposeBase, ComposeOptional, Mix, Optional


def test_field():
    assert id(Base.StringField) != id(Optional.OptString)


def test_required_dict():
    assert id(ComposeBase.FirstBase) != id(ComposeBase.SecondBase)


def test_optional_dict():
    assert id(ComposeOptional.FirstOpt) != id(ComposeOptional.OptionalOpt)


def test_embedded():
    assert id(Mix.ReqBase) != id(ComposeBase.FirstBase)


def test_field_array():
    assert id(Mix.OptTypeArray) != id(Mix.OptBaseArray)


def test_field_array_generator():
    assert id(Mix.OptTypeArray.field) != id(Mix.OptBaseArray.field)



from tests.data import Mix


mix = Mix()
mix_full = Mix(is_full=True)


def test_generated_dict_is_required():
    assert mix_full.ReqOptional.required is Mix.ReqOptional.required


def test_generated_embedded_dict_is_required():
    assert (Mix.OptOptional.OptString.required is
            mix.OptOptional.OptString.required)


def test_generated_embedded_optional_dict_is_required():
    assert Mix.OptBase.IntField.required is mix_full.OptBase.IntField.required


def test_generated_array_is_required():
    assert Mix.OptFullBaseArray.required is mix_full.OptFullBaseArray.required


def test_generated_embedded_array_dict_is_required():
    assert (Mix.OptBaseArray.field.StringField.required is
            mix_full.OptBaseArray[0].StringField.required)




class BaseEnum:
    @classmethod
    def _attr_values_as_list(cls):
        result = []

        for attr_name in dir(cls):
            if not attr_name.startswith('_'):
                attr_value = getattr(cls, attr_name)

                if (not callable(attr_value) and
                        not isinstance(attr_value, property)):
                    result.append(attr_value)

        return result

    @classmethod
    def to_list(cls, exclude=()):
        """
        :param exclude: list of values to be excluded from result; if
          not existent value passed, ValueError exception will be thrown
        :return: list of all fields of class A(Enum) and its parents
        """
        result = cls._attr_values_as_list()
        if type(exclude) not in (list, tuple):
            exclude = (exclude,)

        for item in exclude:
            result.remove(item)

        return result



import decimal
import random

from datetime import datetime
from dateutil import tz

from surety.sdk import fakeable, dates
from surety.sdk.base_enum import BaseEnum
from surety.sdk.field import Field
from surety.process import dictionary


class Bool(Field):
    def generate_value(self):
        return fakeable.FAKER.boolean()


class Int(Field):
    def __init__(self, name=None, required=True, allow_none=False,
                 min_val=None, max_val=None, default=None):
        self.save_kwargs(locals())
        self.min_val = min_val or 0
        self.max_val = max_val or 9999
        super().__init__(
            name=name, required=required, allow_none=allow_none,
            default=default
        )

    def generate_value(self):
        return fakeable.FAKER.pyint(
            min_value=self.min_val,
            max_value=self.max_val
        )


class Float(Field):
    def __init__(self, name=None, required=True, allow_none=False, i_len=None,
                 f_len=None, min_val=None, max_val=None, positive=None,
                 default=None):
        self.save_kwargs(locals())
        self.i_len = i_len
        self.f_len = f_len
        self.min_val = min_val
        self.max_val = max_val
        self.positive = positive
        super().__init__(
            name=name, required=required, allow_none=allow_none,
            default=default
        )

    def generate_value(self):
        return fakeable.FAKER.pyfloat(
            left_digits=self.i_len,
            right_digits=self.f_len,
            positive=bool(self.positive),
            min_value=self.min_val,
            max_value=self.max_val
        )


class Decimal(Float):
    def generate_value(self):
        return fakeable.FAKER.pydecimal(
            left_digits=self.i_len or 1,
            right_digits=self.f_len,
            positive=self.positive is None or self.positive,
            min_value=self.min_val,
            max_value=self.max_val
        )

    @staticmethod
    def from_value(value):
        if isinstance(value, (str, float)):
            value = decimal.Decimal(str(value))

        return Decimal().with_values(value)


class StringDecimal(Decimal):
    @property
    def value(self):
        if self._value in [None, '']:
            return self._value

        return f'{float(self._value):.{self.f_len or 2}f}'


class String(Field):
    def __init__(self, name=None, required=True, min_len=1,
                 max_len=None, allow_none=False, default=None, fake_as=None):
        self.save_kwargs(locals())
        self.min_len = min_len
        self.max_len = max_len
        self.fake_as = fake_as
        super().__init__(
            name=name, required=required, allow_none=allow_none,
            default=default
        )

    def generate_value(self):
        return fakeable.fake_string_attr(
            self.fake_as or self.name,
            min_len=self.min_len,
            max_len=self.max_len
        )


class Enum(Field, BaseEnum):
    def __init__(self, name=None, required=True, allow_none=False,
                 default=None, exclude=None):
        self.save_kwargs(locals())
        self.exclude = exclude or ()
        super().__init__(
            name=name, required=required, allow_none=allow_none,
            default=default
        )

    def generate_value(self):
        return random.choice(self.__class__.to_list(
            exclude=self.exclude
        ))


class Raw(String):
    def __init__(self, name=None, required=True, allow_none=False):
        self.save_kwargs(locals())
        self.overrided_values = False
        self._value = None
        super().__init__(
            name=name, required=required, allow_none=allow_none
        )

    def generate_value(self):
        return fakeable.FAKER.pydict(10, True, value_types=(str,))

    def with_values(self, values):
        if self.overrided_values and isinstance(self._value, dict):
            self._value = dictionary.merge_with_updates(self._value, values)
        else:
            self._value = values

        self.overrided_values = True

        return self


class Uuid(Field):
    def generate_value(self):
        return fakeable.FAKER.uuid4()


class DateTime(String):
    def __init__(self, name=None, required=True, allow_none=False,
                 default=None,
                 date_format=dates.Pattern.DATETIME_DELIM_T_WITH_ZONE_PRECISED,
                 rstrip_zeros=False):
        self.save_kwargs(locals())
        self.date_format = date_format
        self.rstrip_zeros = rstrip_zeros
        super().__init__(
            name=name, required=required, allow_none=allow_none,
            default=default
        )

    def generate_value(self):
        return dates.generate_time()

    @property
    def value(self):
        if self._value and not isinstance(self._value, str):
            result = self._value.strftime(self.date_format)

            if self.rstrip_zeros:
                while '0Z' in result:
                    result = result.replace('0Z', 'Z')

            return result

        return self._value

    def to_format(self, date_format, new_tz=None):
        if self._value is None:
            return None

        from_zone = tz.tzutc()
        to_zone = tz.gettz(new_tz)
        value = self._value

        if isinstance(value, str):
            value = datetime.strptime(value, self.date_format)

        local_time = value.replace(tzinfo=from_zone)
        new_time = local_time.astimezone(to_zone)

        return new_time.strftime(date_format)



from surety import Array
from tests.data import AllowedNone, Mix, Optional


def test_required_value():
    assert Mix().ReqBase.value


def test_required_embedded_field_value():
    assert Mix().ReqBase.StringField.value


def test_optional_no_value():
    assert Mix().OptBase.value is None


def test_optional_embedded_field_no_value():
    assert Mix().OptBase.IntField.value is None


def test_optional_value_in_full():
    assert Mix(is_full=True).OptBase.value


def test_optional_embedded_field_in_full():
    assert Mix(is_full=True).OptBase.IntField.value


def test_array_value():
    assert Mix().ReqTypeArray.value


def test_set_value():
    assert Mix().ReqTypeSet.value


def test_optional_array_no_value():
    assert Mix().OptBaseArray.value is None


def test_optional_set_no_value():
    assert Mix().OptTypeSet.value is None


def test_optional_array_in_full():
    assert Mix(is_full=True).OptBaseArray.value


def test_optional_set_in_full():
    assert Mix(is_full=True).OptTypeSet.value


def test_optional_field_in_array_no_value():
    assert Array(Optional)[0].OptString.value is None


def test_optional_field_in_array_value_in_full():
    assert Array(Optional, is_full=True)[0].OptString.value


def test_required_dict_with_optional_fields():
    assert Mix().ReqOptional.value == {}


def test_allownone_value_in_dict():
    assert AllowedNone().value == {
        AllowedNone.NoneString.name: None,
        AllowedNone.NoneInt.name: None,
    }



from tests.data import Mix


def test_generated_array_is_full_false():
    mix_full = Mix(is_full=True)
    assert mix_full.OptBaseArray[0].is_full is False


def test_generated_array_is_full_true():
    mix = Mix()
    assert mix.ReqFullBaseArray[0].is_full



import json

from surety.sdk.array import Array
from surety.sdk.field import Field


class Dictionary(Field):
    def __init__(self, name=None, required=True, allow_none=False,
                 is_full=False):
        self._generated = False
        self._is_none = False
        self.save_kwargs(locals())
        self.is_full = is_full
        super().__init__(name, required, allow_none)

    def _get_field_names(self):
        result = []

        for attr_name in dir(self.__class__):
            if attr_name.startswith('_'):
                continue

            if isinstance(getattr(self.__class__, attr_name), Field):
                result.append(attr_name)

        return result

    @property
    def generated(self):
        if self._generated or (self.required and self.allow_none):
            return True

        for field_name in self._get_field_names():
            field_value = getattr(self, field_name)

            if field_value.generated:
                return True

        return False

    @generated.setter
    def generated(self, generated):
        self._generated = generated

    @property
    def value(self):
        if not self.generated or self._is_none:
            return None

        _value = {}

        for field_name in self._get_field_names():
            field_value = getattr(self, field_name)

            if field_value.generated:
                _value[field_value.name] = field_value.value

        return _value

    @property
    def full_value(self):
        if not self.generated or self._is_none:
            return None

        _value = {}

        for field_name in self._get_field_names():
            field_value = getattr(self, field_name)
            _value[field_value.name] = field_value.full_value

        return _value

    def generate_value(self):
        pass

    def generate_custom(self):
        pass

    def with_values(self, values):
        for field, value in values.items():
            if isinstance(field, Field):
                field = field.name
            self._set_field_value(field, value)

        return self

    def _get_field(self, name):
        for field_name in self._get_field_names():
            field = getattr(self, field_name)

            if field.name == name:
                return field_name, field

        raise AttributeError(f'No attribute with name {name}')

    def _set_field_value(self, field_name, value):
        if value is None:
            return

        self._is_none = False

        if isinstance(value, Field):
            value = value.value

        field_name, field = self._get_field(field_name)

        if isinstance(value, dict):
            new_value = field.with_values(value)

            if isinstance(field, Dictionary):
                new_value.generated = True

            setattr(self, field_name, new_value)

        elif isinstance(value, list) and isinstance(field, Array):
            new_value = []

            for val in value:
                if isinstance(val, Field):
                    val = val.value

                new_value.append(
                    field.field(is_full=self.is_full).with_values(val)
                )

            setattr(self, field_name, new_value)
        else:
            setattr(self, field_name, value)

    def _generate(self, is_full=True, with_data=True, required=True,
                  use_default=True):
        self._generated = with_data

        if self.allow_none:
            if not is_full:
                self._is_none = True
                return  # Do not generate fields if None is allowed

            self._is_none = False

        if not required and not is_full:
            with_data = False
            self._generated = False

        for field_name in self._get_field_names():
            value = getattr(self, field_name)
            _with_data = with_data

            if _with_data:
                if isinstance(value, Field):
                    if is_full:
                        _with_data = True
                    elif value.allow_none:
                        _with_data = False
                    else:
                        _with_data = value.required

            setattr(self, field_name, value(
                is_full=is_full, with_data=_with_data
            ))

        if with_data:
            self.generate_custom()

    def generate(self, is_full=False):  # pylint: disable=arguments-renamed
        self._generate(is_full=is_full)

    def __setattr__(self, key, value):
        if hasattr(self, key):
            current_value = getattr(self, key)
        else:
            current_value = None

        def needs_json_loads(current, update):
            dict_needs_load = (isinstance(current, Dictionary) and
                               not isinstance(update, dict))
            list_needs_load = (isinstance(current, Array) and
                               not isinstance(update, (list, set)))

            return dict_needs_load or list_needs_load

        if (isinstance(current_value, Field) and
                not isinstance(value, Field)):
            if needs_json_loads(current=current_value, update=value):
                try:
                    if isinstance(value, bytes):
                        value = value.decode('utf-8')

                    value = json.loads(value)
                except json.JSONDecodeError as ex:
                    raise AttributeError(
                        'Assigning entity to primitive'
                    ) from ex

            current_value.with_values(value)
            object.__setattr__(self, key, current_value)
        else:
            object.__setattr__(self, key, value)



from tests.data import Mix


mix = Mix()
mix_full = Mix(is_full=True)


def test_generated_dict_name():
    assert mix_full.OptBase.name == Mix.OptBase.name


def test_generated_embedded_dict_name():
    assert mix.ReqOptional.OptInt.name == Mix.ReqOptional.OptInt.name


def test_generated_array_name():
    assert mix.OptFullBaseArray.name == Mix.OptFullBaseArray.name


def test_generated_embedded_array_dict_name():
    assert (mix_full.OptBaseArray[0].StringField.name ==
            Mix.OptBaseArray.field.StringField.name)



from tests.data import Mix


def test_dict_is_required():
    assert Mix.ReqOptional.required


def test_embedded_dict_is_not_required():
    assert not Mix.OptOptional.OptString.required


def test_embedded_optional_dict_is_required():
    assert Mix.OptBase.IntField.required


def test_array_is_not_required():
    assert not Mix.OptFullBaseArray.required


def test_embedded_array_dict_is_required():
    assert Mix.OptBaseArray.field.StringField.required



from tests.data import Mix


def test_dict_name():
    assert Mix.ReqBase.name == 'req_base'


def test_embedded_dict_name():
    assert Mix.ReqOptional.OptString.name == 'opt_string'


def test_array_name():
    assert Mix.OptBaseArray.name == 'opt_base_array'


def test_embedded_array_dict_name():
    assert Mix.OptBaseArray.field.StringField.name == 'string_field'



from tests.data import Base, ComposeBase, Mix


base = Base()
compose_base = ComposeBase()
mix = Mix()
mix_full = Mix(is_full=True)


def test_generated_field():
    assert id(base.StringField.value) != id(
        compose_base.FirstBase.StringField.value)


def test_generated_embedded_field():
    assert id(base.StringField.value) != id(
        compose_base.FirstBase.StringField.value)


def test_generated_array():
    assert mix.ReqBaseArray.value != mix_full.ReqFullBaseArray.value


def test_array_contract():
    assert id(mix.ReqTypeArray.field.value) != id(mix.ReqTypeArray[0].value)


def test_set_contract():
    assert id(mix.ReqTypeSet.field.value) != id(mix.ReqTypeSet[0].value)



from surety import Array
from tests.data import Mix, Optional


def test_required_full_value():
    assert Mix().ReqBase.full_value


def test_required_embedded_field_full_value():
    assert Mix().ReqBase.StringField.full_value


def test_optional_no_full_value():
    assert Mix().OptBase.full_value is None


def test_optional_embedded_field_no_full_value():
    assert Mix().OptBase.IntField.full_value is None


def test_array_full_value():
    assert Mix().ReqTypeArray.full_value


def test_set_full_value():
    assert Mix().ReqTypeSet.full_value


def test_optional_array_no_full_value():
    assert Mix().OptBaseArray.full_value == []


def test_optional_set_no_full_value():
    assert Mix().OptTypeSet.full_value == []


def test_optional_field_in_array_no_full_value():
    assert Array(Optional)[0].OptString.full_value is None


def test_required_dict_with_optional_fields_full_value():
    assert Mix().ReqOptional.full_value == {
        Mix.ReqOptional.OptInt.name: None,
        Mix.ReqOptional.OptString.name: None,
    }



from tests.data import AllowedNone, ComposeOptional, Mix, Optional


def test_generate_field_is_generated():
    entity = Optional()
    entity.OptInt.generate()
    assert entity.OptInt.generated


def test_generate_field_value():
    entity = Optional()
    entity.OptInt.generate()
    assert entity.OptInt.value


def test_generate_field_not_affect_other_generated():
    entity = Optional()
    entity.OptInt.generate()
    assert entity.OptString.generated is False


def test_generated_field_not_affect_other_values():
    entity = Optional()
    entity.OptInt.generate()
    assert entity.OptString.value is None


def test_generate_embedded_field_dictionary_is_generated():
    entity = ComposeOptional()
    entity.OptionalOpt.OptString.generate()
    assert entity.OptionalOpt.generated


def test_generate_embedded_field_dictionary_value():
    entity = ComposeOptional()
    entity.OptionalOpt.OptString.generate()
    assert entity.OptionalOpt.value


def test_generate_embedded_field_is_generated():
    entity = ComposeOptional()
    entity.OptionalOpt.OptString.generate()
    assert entity.OptionalOpt.OptString.generated


def test_generate_embedded_field_value():
    entity = ComposeOptional()
    entity.OptionalOpt.OptString.generate()
    assert entity.OptionalOpt.OptString.value


def test_generate_embedded_field_not_affect_other_is_generated():
    entity = ComposeOptional()
    entity.OptionalOpt.OptString.generate()
    assert entity.OptionalOpt.OptInt.generated is False


def test_generate_embedded_field_not_affect_other_value():
    entity = ComposeOptional()
    entity.OptionalOpt.OptString.generate()
    assert entity.OptionalOpt.OptInt.value is None


def test_generate_dictionary_is_generated():
    entity = Mix()
    entity.OptBase.generate()
    assert entity.OptBase.generated


def test_generated_dictionary_value():
    entity = Mix()
    entity.OptBase.generate()
    assert entity.OptBase.value


def test_generate_optional_dictionary_is_generated():
    entity = Mix()
    entity.OptOptional.generate()
    assert entity.OptOptional.generated


def test_generate_optional_dictionary_value():
    entity = Mix()
    entity.OptOptional.generate()
    assert entity.OptOptional.value == {}


def test_generate_allow_none_value():
    entity = AllowedNone()
    entity.NoneString.generate()
    assert entity.NoneString.generated



from tests.data import Base, ComposeBase, Mix


def test_field_not_generated():
    assert not Base.StringField.generated


def test_embedded_dict_not_generated():
    assert not ComposeBase.FirstBase.generated


def test_embedded_dict_field_not_generated():
    assert not ComposeBase.FirstBase.IntField.generated


def test_array_not_generated():
    assert not Mix.ReqFullBaseArray.generated



import calendar
import random

from datetime import datetime, timezone

import pytz


DATE_TIME_PATTERN = '%Y-%m-%dT%H:%M:%SZ'


class Pattern:
    DATE = '%Y-%m-%d'
    TIME = '%H:%M:%S'
    DATETIME = '%Y-%m-%d %H:%M:%S'
    DATETIME_DELIM_T = '%Y-%m-%dT%H:%M:%S'
    DATETIME_DELIM_T_WITH_ZONE = '%Y-%m-%dT%H:%M:%SZ'
    DATETIME_DELIM_T_WITH_ZONE_PRECISED = '%Y-%m-%dT%H:%M:%S.%fZ'


def get_current_time_utc():
    return datetime.now(timezone.utc)


def generate_time():
    now = datetime.now(timezone.utc)

    return now.replace(
        day=random.randint(1, now.day),
        hour=random.randint(0, now.hour),
        minute=random.randint(0, now.minute),
        second=random.randint(0, now.second),
    )


def generate_date():
    time = generate_time()

    return datetime.date(time)


def to_format(date_time, pattern=Pattern.DATETIME_DELIM_T_WITH_ZONE):
    return date_time.strftime(pattern)


def to_timestamp(date_time):
    return int(calendar.timegm(date_time.utctimetuple()))


def str_to_date(date_string, pattern=Pattern.DATETIME_DELIM_T_WITH_ZONE):
    return datetime.strptime(date_string, pattern)


def to_timezone(date_time, tz):
    return date_time.astimezone(pytz.timezone(tz))



import copy

from collections.abc import Callable, Mapping, MutableSequence

from surety.process.common import is_callable_with_strict_args


def merge_with_updates(initial_dict, updates, extend_only=False,
                       merge_lists=False):
    """
    Creates copy of initial_dict with updates applied

    **Parameters**

    initial_dict : dictionary used as base for updates
    updates : dictionary with changes to be applied
    extend_only: boolean flag that prohibits to overwrite values, only add new

    **Returns**

    A copy of initial dict with all updates applied

    **Examples**

    initial_dict is None:
      >>> merge_with_updates(None, {2: 2})
      {2: 2}

    Add new value in dict:
      >>> merge_with_updates({1: 1}, {2: 2})
      {1: 1, 2: 2}

    Update value in dict:
      >>> merge_with_updates({1: 1, 2: 2}, {2: 'two', 3: 3} )
      {1: 1, 2: 'two', 3: 3}

    Update value in inner dict:
      >>> merge_with_updates({1: 1, 2: {21: 21}}, {2: {21: 'twenty-one'}})
      {1: 1, 2: {21: 'twenty-one'}}

    Override list:
      >>> merge_with_updates({1: [{2: 2, 3: 3}]}, {1: [{2: 'two'}]})
      {1: [{2: 'two'}]}

    Update value in list:
      >>> merge_with_updates(
      ...   {1: [{2: 2, 3: 3}]}, {1: [{2: 'two'}]},
      ...   merge_lists=True)
      {1: [{2: 'two', 3: 3}]}

    Update value in inner list:
      >>> merge_with_updates(
      ...   {1: {2: [{3: [{4: 4}]}]}},
      ...   {1: {2: [{3: [{4: 'four'}]}]}},
      ...   merge_lists=True)
      {1: {2: [{3: [{4: 'four'}]}]}}

    Extend dict:
      >>> merge_with_updates({1: 1}, {2: 2}, extend_only=True)
      {1: 1, 2: 2}

    Extend list:
      >>> merge_with_updates({1: [2]}, {1: [2, 3]}, merge_lists=True)
      {1: [2, 2, 3]}

    Extend list of dicts:
      >>> merge_with_updates({1: [{2: 2}]}, {1: [{3: 3}]}, merge_lists=True)
      {1: [{2: 2, 3: 3}]}

    Extend empty list with list of dicts:
      >>> merge_with_updates({1: []}, {1: [{2: 2}]}, merge_lists=True)
      {1: [{2: 2}]}

    Extend list of dicts with empty inner dict:
      >>> merge_with_updates({1: [{2: 2}]}, {1: [{}]}, merge_lists=True)
      {1: [{2: 2}, {}]}

    Extend inner dict:
      >>> merge_with_updates({1: 1, 2: {21: 21}}, {2: {22: 22}})
      {1: 1, 2: {21: 21, 22: 22}}

    Overwrite value with dict:
      >>> merge_with_updates({1: 1}, {1: {2: 2}})
      {1: {2: 2}}

    Do not modify initial dict:
      >>> initial_dict = {1: 1, 2: {21: 21}}
      >>> merge_with_updates(initial_dict, {1: 'one', 2: {21: 'two'}})
      {1: 'one', 2: {21: 'two'}}
      >>> print(initial_dict)
      {1: 1, 2: {21: 21}}

    Do not override value in extend mode:
      >>> merge_with_updates({1: 1}, {1: 'one'}, extend_only=True)
      Traceback (most recent call last):
          ...
      ValueError: Can not overwrite "1" value in extend mode

    Do not override inner value in extend mode:
      >>> merge_with_updates({1: {2: 2}}, {1: {2: 0}}, extend_only=True)
      Traceback (most recent call last):
          ...
      ValueError: Can not overwrite "2" value in extend mode

    Do not override list value in extend mode:
      >>> merge_with_updates(
      ...   {1: [{2: 2}]}, {1: [{2: 0}]}, merge_lists=True, extend_only=True)
      Traceback (most recent call last):
          ...
      ValueError: Can not overwrite "2" value in extend mode
    """
    initial_dict = initial_dict or {}
    initial_copy = copy.deepcopy(initial_dict)

    for key, value in updates.items():
        if isinstance(value, Mapping) and isinstance(initial_copy.get(key),
                                                     Mapping):
            initial_copy[key] = merge_with_updates(
                initial_copy.get(key, {}),
                value,
                extend_only=extend_only,
                merge_lists=merge_lists
            )
        elif merge_lists and isinstance(value, list) and isinstance(
                initial_copy.get(key), list):
            if len(initial_copy[key]) == 0 or not isinstance(
                    value[0], Mapping) or value[0] == {}:
                initial_copy[key].extend(value)
            else:
                initial_copy[key][0] = merge_with_updates(
                    initial_copy.get(key, [{}])[0],
                    value[0],
                    extend_only=extend_only,
                    merge_lists=merge_lists
                )
        else:
            if extend_only and key in initial_copy:
                raise ValueError(
                    f'Can not overwrite "{key}" value in extend mode'
                )

            initial_copy[key] = updates[key]

    return initial_copy


def dict_predicate(func=None, name=None):
    class DictPredicate:
        def __init__(self, callable_func):
            self.func = callable_func

        def __call__(self, *args, **kwargs):
            return self.func(*args, **kwargs)

        def __repr__(self):
            return name or self.func.__name__

    if func:
        return DictPredicate(func)

    return DictPredicate


def matches_pattern(initial_dict, pattern_dict):
    """
    Checks whether dict matches value/filter pattern

    **Parameters**

    initial_dict : dictionary to verify
    pattern_dict: dictionary pattern to apply for verification. Filter function
    is allowed as value: takes one actual arg and returns True or False.
    Decorate filter functions with @dict_predicate to add names on logging.

    **Returns**

    True/False if the dict matches pattern.

    **Examples**

    Full match:
      >>> matches_pattern({1: 1}, {1: 1})
      True

    Extra value in initial dict:
      >>> matches_pattern({1: 1, 2: 2}, {2: 2})
      True

    Extra value in pattern:
      >>> matches_pattern({1: 1}, {1: 1, 2: 2})
      False

    Inner dict doesn't match filter pattern:
      >>> matches_pattern({1: {2: 2}}, {1: {2: lambda x: x != 2}})
      False

    Inner dict matches filter pattern:
      >>> matches_pattern({1: {2: 2}}, {1: {2: lambda x: x == 2}})
      True
    """
    result = True

    for key, value in pattern_dict.items():
        if result is False or key not in initial_dict:
            return False

        if isinstance(value, Mapping):
            result = matches_pattern(initial_dict[key], value)
        elif is_callable_with_strict_args(value):
            result = value(initial_dict[key])
        else:
            result = value == initial_dict[key]

    return result


def apply_modifier(initial_dict, modifier):
    """
    Returns a copy of an initial dictionary with modified values.

    :param initial_dict: any dictionary to create a copy from and modify
    :param modifier: either a callable that modifies the dictionary or a
    dictionary that contains other modifiers under corresponding keys.
    :return: a copy of the dictionary
    :rtype: dict

    Examples:
      >>> apply_modifier({'a': 7}, lambda _: {})
      {}

      >>> apply_modifier({'b': 7}, {'b': float})
      {'b': 7.0}

      >>> apply_modifier({'c': {'x': 1}, 'd': 2}, {'c': {'x': float}})
      {'c': {'x': 1.0}, 'd': 2}

      >>> apply_modifier({'c': {'x': 1}}, {'d': float})
      {'c': {'x': 1}}

      >>> apply_modifier({'c': {'x': 1}}, {'c': lambda _: 'test'})
      {'c': 'test'}

      >>> apply_modifier({'c': [1, 2, 3]}, {'c': [float]})
      {'c': [1.0, 2.0, 3.0]}

      >>> apply_modifier({'c': []}, {'c': [float]})
      {'c': []}

      >>> apply_modifier(
      ...   {'c': [{'x': [1]}, {'y': 4}]}, {'c': [{'x': [float]}]})
      {'c': [{'x': [1.0]}, {'y': 4}]}

      >>> apply_modifier(
      ...   {'c': [{'x': [1], 'y': 2}]}, {'c': [{'x': [float], 'y': float}]})
      {'c': [{'x': [1.0], 'y': 2.0}]}

    """
    initial_copy = copy.deepcopy(initial_dict)

    if isinstance(modifier, Callable):
        return modifier(initial_copy)

    if isinstance(modifier, MutableSequence):
        if not modifier:
            raise ValueError('list modifier is empty')

        if not isinstance(initial_copy, MutableSequence):
            raise ValueError(
                f'cannot apply list modifier to {type(initial_copy)} type'
            )
        sub_modifier = modifier[0]

        return [apply_modifier(v, sub_modifier) for v in initial_copy]

    if not isinstance(modifier, Mapping):
        raise ValueError('modifier is not dictionary nor callable')

    for key, sub_modifier in modifier.items():
        if key not in initial_copy:
            continue

        initial_copy[key] = apply_modifier(initial_copy[key], sub_modifier)

    return initial_copy

# pylint: disable=line-too-long
def normalize(actual, expected, keys=None):
    """
    Applies same structure as expected, based on id for ordered structures.

    Examples:
      >>> normalize(['c', 'b', 'a'], ['b'])
      ['b', 'c', 'a']

      >>> normalize({'a': [1, 2, 5, 6], 'b': 0}, {'a': [1, 3, 6 ], 'b': None})
      {'a': [1, 6, 2, 5], 'b': 0}

      >>> normalize(
      ...   {'a': {'c': 'test', 'd': ['a', 1]}, 'b': 123},
      ...   {'a': {'d': ['b', 'c', 1]}})
      {'a': {'c': 'test', 'd': [1, 'a']}, 'b': 123}

      >>> normalize(
      ... {'a': [{'id': 1, 'val': 'test'}, {'id': 2, 'val': 'test2'}]},
      ... {'a': [{'id': 2}, {'id': 3}]})
      {'a': [{'id': 2, 'val': 'test2'}, {'id': 1, 'val': 'test'}]}

      >>> normalize(
      ...   {'a': [{'id': 1, 'val': 'test'}, {'id': 2, 'val': 'test2'}]},
      ...   {'a': [{'id': 4, 'val': 'test2'}, {'id': 5, 'val': 'v'}]},
      ...   keys=['fake', 'val'])
      {'a': [{'id': 2, 'val': 'test2'}, {'id': 1, 'val': 'test'}]}

      >>> normalize(
      ...   {'a': [
      ...       {'id': 'b1', 'b': [{'id': 2}, {'id': 1}]},
      ...       {'id': 'b0', 'b': [{'id': 3}, {'id': 4}]}
      ...   ]},
      ...   {'a': [
      ...       {'id': 'b0', 'b': [{'id': 4}]},
      ...       {'id': 'b1', 'b': [{'id': 1}, {'id': 2}]}
      ...   ]})
      {'a': [{'id': 'b0', 'b': [{'id': 4}, {'id': 3}]}, {'id': 'b1', 'b': [{'id': 1}, {'id': 2}]}]}

      >>> normalize(
      ...   [{
      ...       'attempt_number': 0,
      ...       'cost': 280.333,
      ...       'data_source_name':
      ...       '_all_datasources_',
      ...       'data_transferred_bytes': 9046,
      ...       'job_id': '6858248'
      ...   }, {
      ...       'attempt_number': 0,
      ...       'cost': 280.333,
      ...       'data_source_name': 'azure_blob',
      ...       'data_transferred_bytes': 9046,
      ...       'job_id': '6858248'
      ...   }],
      ...   [{
      ...       'attempt_number': 0,
      ...       'cost': 280.333,
      ...       'data_source_name': 'azure_blob',
      ...       'data_transferred_bytes': 9046,
      ...       'job_id': '6858248'
      ...   }, {
      ...       'attempt_number': 0,
      ...       'cost': 280.333,
      ...       'data_source_name': '_all_datasources_',
      ...       'data_transferred_bytes': 9046,
      ...       'job_id': '6858248'
      ...   }],
      ...   keys=['data_source_name', 'cost'])
      [{'attempt_number': 0, 'cost': 280.333, 'data_source_name': 'azure_blob', 'data_transferred_bytes': 9046, 'job_id': '6858248'}, {'attempt_number': 0, 'cost': 280.333, 'data_source_name': '_all_datasources_', 'data_transferred_bytes': 9046, 'job_id': '6858248'}]
    """
    # pylint: enable=line-too-long
    keys = keys or []

    def find_keys(element):
        if not isinstance(element, dict):
            return []

        result = []

        for available_key in keys:
            if available_key in element:
                result.append(available_key)

        return result or ['id']

    def get_index(tar, elem, el_keys=None):
        el_keys = el_keys or ['id']

        def get_combined_key(values_dict):
            return '--'.join(f'{k}::{values_dict.get(k)}' for k in el_keys)

        target = tar
        element = elem

        if isinstance(elem, dict):

            target = [get_combined_key(item) for item in tar]
            element = get_combined_key(elem)

        if element not in target:
            return len(target)

        return target.index(element)

    initial = copy.deepcopy(actual)

    if isinstance(initial, list) and expected is not None:
        element_keys = find_keys(initial[0]) if initial else None
        initial = sorted(
            initial,
            key=lambda x: get_index(expected, x, el_keys=element_keys)
        )
        if len(initial) == len(expected):
            for num, _ in enumerate(initial):
                initial[num] = normalize(
                    initial[num], expected[num], keys=keys
                )

    elif isinstance(initial, dict):
        for key, value in initial.items():
            if expected:
                initial[key] = normalize(value, expected.get(key), keys=keys)

    return initial



import decimal
import uuid

from datetime import datetime

from surety import (
    Bool, DateTime, Decimal, Enum, Float, Int, Raw, String, StringDecimal,
    Uuid, Dictionary
)


def test_generate_bool():
    value = Bool().value
    assert isinstance(value, bool)


def test_generate_int():
    value = Int().value
    assert isinstance(value, int)


def test_generate_float():
    value = Float().value
    assert isinstance(value, float)


def test_generate_decimal():
    value = Decimal().value
    assert isinstance(value, decimal.Decimal)


def test_decimal_from_str_value():
    value = Decimal.from_value('2.02').value
    assert isinstance(value, decimal.Decimal)


def test_decimal_from_float_value():
    value = Decimal.from_value(1.9).value
    assert isinstance(value, decimal.Decimal)


def test_generate_string_decimal():
    value = StringDecimal().value
    assert isinstance(value, str)


def test_generate_string_decimal_value():
    value = StringDecimal().value
    assert isinstance(decimal.Decimal(value), decimal.Decimal)


def test_string_decimal_no_value():
    value = StringDecimal().with_values('').value
    assert value == ''


def test_generate_string():
    value = String().value
    assert isinstance(value, str)


def test_generate_enum():
    class EnumVals(Enum):
        Example = 'example'

    assert EnumVals().value == 'example'


def test_enum_to_list():
    class EnumVals(Enum):
        Value1 = 'val_1'
        Value2 = 'val_2'
        Value3 = 'val_3'

    assert EnumVals.to_list() == ['val_1', 'val_2', 'val_3']


def test_enum_to_list_exclude_value():
    class EnumVals(Enum):
        Value1 = 'val_1'
        Value2 = 'val_2'
        Value3 = 'val_3'

    assert EnumVals.to_list(exclude=EnumVals.Value2) == ['val_1', 'val_3']


def test_enum_to_list_exclude_multiple_values():
    class EnumVals(Enum):
        Value1 = 'val_1'
        Value2 = 'val_2'
        Value3 = 'val_3'

    assert EnumVals.to_list(
        exclude=(EnumVals.Value2, EnumVals.Value3)
    ) == ['val_1']


def test_generate_raw():
    value = Raw().value
    assert isinstance(value, dict)


def test_raw_with_values():
    value = Raw().with_values({'test': 123}).value
    assert value == {'test': 123}


def test_raw_with_values_override():
    value = Raw().with_values({
        'first': 'val1',
        'second': 'val2',
        'third': 'val3'
    }).with_values({'second': 'val_two'}).value
    assert value == {
        'first': 'val1',
        'second': 'val_two',
        'third': 'val3'
    }


def test_generate_datetime_type():
    value = DateTime().value
    assert isinstance(value, str)


def test_generate_datetime_type_trailing_zeroes():
    expected = '2026-02-10T09:07:11.404000Z'
    value = DateTime().with_values(expected).value
    assert value == expected


def test_generate_datetime_type_rstrip_trailing_zeroes():
    initial = '2026-02-10T09:07:11.404000Z'
    expected = '2026-02-10T09:07:11.404Z'
    set_value = datetime.strptime(initial, '%Y-%m-%dT%H:%M:%S.%fZ')
    value = DateTime(rstrip_zeros=True).with_values(set_value).value
    assert value == expected

def test_datetime_to_format_no_value():
    class OptionalDateTime(Dictionary):
        DateTime = DateTime(name='date_time', required=False)

    value = OptionalDateTime().DateTime.to_format('%Y-%m-%dT%H:%M:%S.%fZ')
    assert value is None

def test_datetime_to_format():
    initial = '2026-02-10T09:07:11.404000Z'
    expected = '2026-10-02'
    set_value = datetime.strptime(initial, '%Y-%m-%dT%H:%M:%S.%fZ')
    value = DateTime().with_values(set_value).to_format('%Y-%d-%m')
    assert value == expected


def test_datetime_to_format_from_string():
    initial = '2026-02-10T09:07:11.404000Z'
    expected = '09:07:11'
    value = DateTime().with_values(initial).to_format('%H:%M:%S', new_tz='utc')

    assert value == expected


def test_generate_uuid_type():
    value = Uuid().value
    assert isinstance(value, str)


def test_generate_uuid_value():
    value = Uuid().value
    assert isinstance(uuid.UUID(value, version=4), uuid.UUID)
